//
//  Lexical.m
//  Calc
//
//  Created by KOYAMA Yoshiaki on 05/05/31.
//  Copyright 2005 KOYAMA Yoshiaki. All rights reserved.
//

#import "Lexical.h"
#import "ErrorMessage.h"

/*
	typedef unsigned short unichar; // 2byte
*/

@implementation Lexical

- (id)init {
    self = [super init];
	[self initTable];

    return self;
}

- (void)initTable {
	int i;

	for (i=0; i<0xffff; i++) {
		_table[i] = tOthers;
	}

	for (i='0'; i<='9'; i++) {
		_table[i] = tDigit;
	}

	for (i='a'; i<='z'; i++) {
		_table[i] = tLetter;
	}

	for (i='A'; i<='Z'; i++) {
		_table[i] = tLetter;
	}

	_table['+']  = tPlus;
	_table['-']  = tMinus;
	_table['*']  = tMult;
	_table['/']  = tDiv;
	_table['(']  = tLparn;
	_table[')']  = tRparn;
	_table['^']  = tPow;
		
	_table[' ']  = tSpace;
	_table['\t'] = tTab;
	_table['\n'] = tRet;
	_table['.']  = tPeriod;
	
	_index = 0;
}

- (void)setText:(NSString *)text {
	_text = text;
	[_text retain];
}

- (NSString *)text {
	return _text;
}

- (void)setErrorMessage:(ErrorMessage *)errorMessage {
	_errorMessage = errorMessage;
	[_errorMessage retain];
}

- (void)analyze {
//	unichar charString;
	unsigned int length;
//	int i;

	length = [_text length];

	NSLog(@"%@", _text);
/*	for (i=0; i<length; i++) {
		charString = [_text characterAtIndex:i];
		NSLog(@"%d : %x", i, charString);
	}
*/
//	NSLog(@"%d", sizeof(unichar));

	token temp;
	while (_index < length) {
		temp = [self makeToken];

		NSLog(@"kind: %d value: %f", temp.kind, temp.value);
	}
}

- (token)makeToken {
	token temp;
	unichar charString;
	unsigned int length;
	int kind;
	int index1;

	temp.kind = 0;
	temp.value = 0;


	length = [_text length];

//NSLog(@"topkind %d index %d length %d", kind, _index, length);

	if (_index >= length) {
		return temp;
	}


	charString = [_text characterAtIndex:_index];
	kind = _table[charString];


	_index++;


	while ((kind == tSpace || kind == tTab || kind == tRet) && (_index < length)){
		charString = [_text characterAtIndex:_index];
		kind = _table[charString];
		_index++;
	} 

	double num = 0;
	NSString *sub;
	NSRange aRange;
	
	switch (kind) {
	case tDigit:
		index1 = _index - 1;
		do {
			if (_index < length) {
				charString = [_text characterAtIndex:_index];
				kind = _table[charString];
			}
			_index++;
		} while (((kind  == tDigit) || (kind == tPeriod)) && (_index <= length));
		--_index;
		aRange.location = index1;
		aRange.length = _index - index1;
		sub = [_text substringWithRange:aRange];

		num = [sub doubleValue];
//		NSLog(@"sub:%@", sub);
//	NSLog(@"num  %d index %d length %d", num, _index, length);
		temp.kind = tNum;
		temp.value = num;
		break;
	case tLetter:
		index1 = _index - 1;
		do {
			if (_index < length) {
				charString = [_text characterAtIndex:_index];
				kind = _table[charString];
			}
			_index++;
		} while ((kind  == tLetter) && (_index <= length));
		--_index;
		aRange.location = index1;
		aRange.length = _index - index1;
		sub = [_text substringWithRange:aRange];

		if ([sub caseInsensitiveCompare:@"sin"] == NSOrderedSame) {
			temp.kind = tSin;
		} else if ([sub caseInsensitiveCompare:@"cos"] == NSOrderedSame) {
			temp.kind = tCos;
		} else if ([sub caseInsensitiveCompare:@"tan"] == NSOrderedSame) {
			temp.kind = tTan;
		} else if ([sub caseInsensitiveCompare:@"pi"] == NSOrderedSame) {
			temp.kind = tNum;
			temp.value = M_PI;
		} else if ([sub caseInsensitiveCompare:@"sqrt"] == NSOrderedSame) {
			temp.kind = tSqrt;
		} else if ([sub caseInsensitiveCompare:@"log"] == NSOrderedSame) {
			temp.kind = tLog;
		} else if ([sub caseInsensitiveCompare:@"ln"] == NSOrderedSame) {
			temp.kind = tLn;
		} else if ([sub caseInsensitiveCompare:@"exp"] == NSOrderedSame) {
			temp.kind = tExp;
		} else if ([sub caseInsensitiveCompare:@"e"] == NSOrderedSame) {
			temp.kind = tNum;
			temp.value = M_E;
		} else {
			temp.kind = tOtherFunc;
			[_errorMessage setErrorFlag:YES];
			[_errorMessage addNumber:ERROR_NO_FUNCTION];
		}
//		NSLog(@"sub:%@", sub);
//		NSLog(@"kind  %d index %d length %d", temp.kind, _index, length);
		break;
	case tPlus:
		temp.kind = tPlus;
		break;
	case tMinus:
		temp.kind = tMinus;
		break;
	case tMult:
		temp.kind = tMult;
		break;
	case tDiv:
		temp.kind = tDiv;
		break;
	case tLparn:
		temp.kind = tLparn;
		break;
	case tRparn:
		temp.kind = tRparn;
		break;
	case tPow:
		temp.kind = tPow;
		break;
	case tOthers:
		[_errorMessage setErrorFlag:YES];
		[_errorMessage addNumber:ERROR_OTHERS];
		temp.kind = tOthers;
		break;
	}

	return temp;
}


- (void)dealloc {
	[_errorMessage release];
	[_text release];
    [super dealloc];
}

@end
