//
//  Syntax.m
//  Calc
//
//  Created by KOYAMA Yoshiaki on 05/06/01.
//  Copyright 2005 KOYAMA Yoshiaki. All rights reserved.
//

#import "Syntax.h"
#import "Lexical.h"
#import "ErrorMessage.h"

#include <math.h>

@implementation Syntax

- (id)init {
    self = [super init];

	_spCalc = 0;

    return self;
}

- (void)setLex:(Lexical *)lex {
	_lex = lex;
	[_lex retain];
}

- (void)setErrorMessage:(ErrorMessage *)errorMessage {
	_errorMessage = errorMessage;
	[_errorMessage retain];
}

- (double)analyze {
	NSString *text;

	text = [_lex text];

#ifdef DEBUG_CALC
	NSLog(@"-------------------------------");
	NSLog(@"%@", text);
	NSLog(@"-------------------------------");
#endif	
	_token = [_lex makeToken];
	[self expression];

#ifdef DEBUG_CALC
	NSLog(@"stack %f", _stackCalc[_spCalc-1]);
#endif
	return _stackCalc[_spCalc-1];
}

/*
	 (expression)
*/
- (void)expression {
	int kind;
	double t1, t2;

	[self term];
	while ((_token.kind == tPlus) || (_token.kind == tMinus)) {
		kind = _token.kind;
		_token = [_lex makeToken];		
		[self term];

		t1 = _stackCalc[_spCalc-2];
		t2 = _stackCalc[_spCalc-1];

		if (kind == tPlus) { 
			_stackCalc[_spCalc-2] = t1 + t2;
#ifdef DEBUG_CALC
			NSLog(@"+");
#endif
		} else {
			_stackCalc[_spCalc-2] = t1 - t2;
#ifdef DEBUG_CALC
			NSLog(@"-");
#endif
		}

		_spCalc = _spCalc - 1;
	}
}

/*
	 (term)
*/

- (void)term {
	int kind;
	double t1, t2;

	[self factor];
	while ((_token.kind == tMult) || (_token.kind == tDiv) || (_token.kind == tPow)) {
		kind = _token.kind;
		_token = [_lex makeToken];
		[self factor];

		t1 = _stackCalc[_spCalc-2];
		t2 = _stackCalc[_spCalc-1];

		if (kind == tMult) {
			_stackCalc[_spCalc-2] = t1 * t2;
#ifdef DEBUG_CALC
			NSLog(@"*");
#endif
		} else if (kind == tDiv) {
			_stackCalc[_spCalc-2] = t1 / t2;
#ifdef DEBUG_CALC
			NSLog(@"/");
#endif
		} else if (kind == tPow) {
			_stackCalc[_spCalc-2] = pow(t1, t2);
#ifdef DEBUG_CALC
			NSLog(@"^");
#endif
		}

		_spCalc = _spCalc - 1;
	}

}

/*
	q (factor)
*/
- (void)factor {
	int kind;
	double t1;

	if (_token.kind == tLparn) {
		_token = [_lex makeToken];

		[self expression];

		if (_token.kind == tRparn) {
#ifdef DEBUG_CALC
			NSLog(@"Rparn");
#endif
		} else {
			[_errorMessage setErrorFlag:YES];
			[_errorMessage addNumber:ERROR_NO_RPARN];
#ifdef DEBUG_CALC
			NSLog(@"no find Rparn");
#endif
		}
		_token = [_lex makeToken];

	} else if ((_token.kind == tSin) || (_token.kind == tCos) || (_token.kind == tTan) ||
				(_token.kind == tSqrt) || (_token.kind == tLog) || (_token.kind == tLn) || 
				(_token.kind == tExp)) {
		kind = _token.kind;

		_token = [_lex makeToken];

		if (_token.kind == tLparn) {
			_token = [_lex makeToken];
		} else {
			[_errorMessage setErrorFlag:YES];
			[_errorMessage addNumber:ERROR_NO_LPARN];
		}

		[self expression];

		if (_token.kind == tRparn) {
#ifdef DEBUG_CALC
			NSLog(@"Rparn");
#endif
		} else {		
			[_errorMessage setErrorFlag:YES];
			[_errorMessage addNumber:ERROR_NO_RPARN];
#ifdef DEBUG_CALC
			NSLog(@"no find Rparn");
#endif
		}

		t1 = _stackCalc[_spCalc-1];

		if (kind == tSin) {
			_stackCalc[_spCalc-1] = sin(t1);
#ifdef DEBUG_CALC
			NSLog(@"sin(%f) = %f", t1, _stackCalc[_spCalc-1]);
#endif
		} else if (kind == tCos) {
			_stackCalc[_spCalc-1] = cos(t1);
#ifdef DEBUG_CALC
			NSLog(@"cos(%f) = %f", t1, _stackCalc[_spCalc-1]);
#endif
		} else if (kind == tTan) {		
			_stackCalc[_spCalc-1] = tan(t1);
#ifdef DEBUG_CALC
			NSLog(@"tan(%f) = %f", t1, _stackCalc[_spCalc-1]);
#endif
		} else if (kind == tSqrt) {		
			_stackCalc[_spCalc-1] = sqrt(t1);
#ifdef DEBUG_CALC
			NSLog(@"sqrt(%f) = %f", t1, _stackCalc[_spCalc-1]);
#endif
		} else if (kind == tLog) {		
			_stackCalc[_spCalc-1] = log10(t1);
#ifdef DEBUG_CALC
			NSLog(@"log(%f) = %f", t1, _stackCalc[_spCalc-1]);
#endif
		} else if (kind == tLn) {		
			_stackCalc[_spCalc-1] = log(t1);
#ifdef DEBUG_CALC
			NSLog(@"ln(%f) = %f", t1, _stackCalc[_spCalc-1]);
#endif
		} else if (kind == tExp) {		
			_stackCalc[_spCalc-1] = exp(t1);
#ifdef DEBUG_CALC
			NSLog(@"exp(%f) = %f", t1, _stackCalc[_spCalc-1]);
#endif
		}
				
		_token = [_lex makeToken];

	} else if (_token.kind == tNum) {

#ifdef DEBUG_CALC	
		NSLog(@"%f", _token.value);
#endif
		_stackCalc[_spCalc] = _token.value;
		_spCalc = _spCalc + 1;
		
		_token = [_lex makeToken];

//		[self printToken];
	} else if ((_token.kind == tPlus) || (_token.kind == tMinus)) {
		kind = _token.kind;

		_token = [_lex makeToken];
		[self factor];

		t1 = _stackCalc[_spCalc-1];
		if (kind == tPlus) {
#ifdef DEBUG_CALC
			NSLog(@"(+)");		
#endif
		} else {
		_stackCalc[_spCalc-1] = -t1;
#ifdef DEBUG_CALC
			NSLog(@"(-)");	
#endif
		}
	} else {
		[_errorMessage setErrorFlag:YES];
		[_errorMessage addNumber:ERROR_NO_FACTOR];
	}
}

- (void)printToken {
	NSLog(@"token: kind %d value %f", _token.kind, _token.value);
}

- (void)dealloc {
	[_errorMessage release];
	[_lex release];
    [super dealloc];
}

@end
